# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.BlockStates do
  use GenServer
  @moduledoc """
  GenServer which can be populated with block states and their corresponding IDs.
  It can be queried by block identifier and state to get the corresponding ID.
  """
  require Logger

  def start_link(map, name) do
    GenServer.start_link(__MODULE__, map, name: name)
  end

  def init(map) do
    {:ok, map}
  end

  def handle_cast({:add, id, bs, bsi}, state) do
    state = Map.put_new(state, id, %{})
    state = Map.put(state, id, Map.put(state[id], bs, bsi))
    {:noreply, state}
  end

  def handle_call({:get, id, bs}, _from, state) do
    case Map.get(state, id) do
      nil -> {:reply, nil, state}
      block_states -> {:reply, Map.get(block_states, bs), state}
    end
  end

  def handle_call(:debug, _from, state) do
    Logger.debug("BlockRegistry state: #{inspect(state)}")
    {:reply, :ok, state}
  end

  @doc """
  Adds a block state to the registry.

  ## Parameters
  - `version` - The Minecraft version
  - `id` - The block identifier
  - `bs` - The block state
  - `bsi` - The block state ID
  """
  @spec add(String.t(), String.t(), map(), integer()) :: :ok
  def add(version, id, bs, bsi) do
    ps = {:via, Registry, {Amethyst.BlockStateRegistry, version}}
    GenServer.cast({:via, PartitionSupervisor, {ps, id}}, {:add, id, bs, bsi})
  end

  @doc """
  Gets the block state ID for a given block identifier and block state.
  If no data is known for this version, errors. Prefer using `try_get/3`
  if you can wait for the data to be loaded.
  - `version` - The Minecraft version
  - `id` - The block identifier
  - `bs` - The block state, nil if the block has no states
  """
  @spec get(String.t(), String.t(), map() | nil) :: integer() | nil
  def get(version, id, bs \\ %{}) do
    ps = {:via, Registry, {Amethyst.BlockStateRegistry, version}}
    GenServer.call({:via, PartitionSupervisor, {ps, id}}, {:get, id, bs})
  end

  @doc """
  Tries to get the block state ID for a given block identifier and block state.
  If no data is known for this version, it will block until the data is loaded.
  - `version` - The Minecraft version
  - `id` - The block identifier
  - `bs` - The block state, nil if the block has no states
  """
  @spec try_get(String.t(), String.t(), map() | nil) :: integer() | nil
  def try_get(version, id, bs \\ %{}) do
    case Registry.lookup(Amethyst.BlockStateRegistry, version) do
      [{_pid, _}] ->
        ps = {:via, Registry, {Amethyst.BlockStateRegistry, version}}
        GenServer.call({:via, PartitionSupervisor, {ps, id}}, {:get, id, bs})
      [] ->
        Logger.warning("BlockStateRegistry doesn't exist for version #{version}, generating it now!")
        Amethyst.DataGenerator.generate_and_populate_data(version, "/tmp/minecraft_server_#{version}.jar", "/tmp/minecraft_data_#{version}")
        try_get(version, id, bs)
    end
  end
end
